import pickle

from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from .assistant import Assistant
from .command_handler import command_handlers, hello_handler, help_handler
from .output_formatter import format_success


class Application:
    def __init__(self):
        pass

    def __load_data(self, filename="storage.bin"):
        with open(filename, "rb") as f:
            return pickle.load(f)

    def __save_data(self, data_dict, filename="storage.bin"):
        with open(filename, "wb") as f:
            pickle.dump(data_dict, f)

    def __parse_command(self, input_sting: str):
        if input_sting == "":
            return "help",

        args = []
        arg = ""
        open_char = None

        for char in input_sting:
            if arg == "":
                if char in ["'", '"', ' '] and open_char is None:
                    open_char = char
                    continue

                if open_char is None:
                    open_char = " "

                arg += char
                continue

            if char == open_char:
                args.append(arg)
                open_char = None
                arg = ""
                continue

            arg += char

        if arg != "":
            args.append(arg)

        command, *arguments = args

        return command, *arguments

    def run(self):
        try:
            assistant = self.__load_data()
        except FileNotFoundError:
            assistant = Assistant()

        print(hello_handler())
        print(help_handler())

        # Предполагаемый список доступных команд
        commands = ['help', 'hello', 'add', 'change', 'phone', 'all', 'delete-contact', 'add-birthday', 'show-birsthday',
                    'birthdays', 'notes', 'add-note', 'edit-note', 'delete-note', 'search-note', 'add-note-tag', 'notes-by-tags',
                    'close', 'exit', 'search-contacts', 'upcoming-birthdays']


        # Создание объекта WordCompleter с доступными командами
        command_completer = WordCompleter(commands, ignore_case=True)

        try:
            while True:
                user_input = prompt('>> ', completer=command_completer)
                command, *arguments = self.__parse_command(user_input)

                if command in ['exit', 'close']:
                    print(format_success('Good bye!'))
                    break

                print(command_handlers(command, assistant, *arguments))

        except KeyboardInterrupt:
            print(format_success('\nGood bye!'))
        finally:
            self.__save_data(assistant)



from collections import UserDict, defaultdict
from typing import List, Dict

from taipan_assistant.notebook.note import Note
from taipan_assistant.search.engine import Engine


class Notebook(UserDict):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.index = 0
        self.search = Engine(["title", "text", "tags"])

    def add_note(self, title: str, note: str) -> str:
        note_id = f'note-{self.index}'
        self.data[note_id] = Note(title, note)
        self.search.add_to_index(note_id, self.data[note_id])
        self.index += 1

        return note_id

    def get_note(self, key: str) -> Note:
        return self.data[key]

    def edit_note(self, note_id: str, new_text) -> None:
        note = self.data[note_id]
        note.edit_note(new_text)
        self.search.add_to_index(note_id, note)

    def delete_note(self, key: str) -> None:
        del self.data[key]
        self.search.remove_from_index(key)

    def find_notes(self, search_text: str) -> List[Note]:
        return self.search.find(search_text)

    def add_tag(self, note_id: str, tag: str) -> None:
        note = self.data[note_id]
        note.add_tag(tag)
        self.search.add_to_index(note_id, note)

    def sort_notes_by_tags(self) -> Dict[str, Dict[str, Note]]:
        tags = defaultdict(dict)
        for note_id, note in self.data.items():
            for tag in note.tags:
                tags[str(tag).lower().capitalize()].update({note_id: note})

        return dict(sorted(tags.items()))



from datetime import datetime

from taipan_assistant.notebook.tag import Tag


class Note:
    def __init__(self, title: str, text: str):
        self.title = title
        self.text = text
        self.tags = set()
        self.updated_at = datetime.now()

    def edit_title(self, title: str):
        self.title = title
        self.updated_at = datetime.now()

    def edit_note(self, text: str):
        self.text = text
        self.updated_at = datetime.now()

    def add_tag(self, tag: str):
        self.tags.add(Tag(tag))
        self.updated_at = datetime.now()

    def __str__(self) -> str:
        return f'''{'-' * 10}
Title: \n{self.title}
{'-' * 10}
Text: \n{self.text}
{'-' * 10}
Tags: \n{', '.join(map(lambda tag: str(tag), self.tags))}
{'-' * 10}
Last changed:
{self.updated_at.strftime("%Y-%m-%d %H:%M:%S")}
'''



from colorama import Fore, Back, Style
from taipan_assistant.assistant.error_decorator import input_error
from taipan_assistant.assistant.output_formatter import format_success, format_error
from taipan_assistant.contacts.address_book import AddressBook
from taipan_assistant.contacts.record.record import Record
from taipan_assistant.notebook.notebook import Notebook


@input_error
def command_handlers(command, assistant, *arguments):
    no_args_handlers_map = {
        'hello': hello_handler,
        'help': help_handler,
    }

    contact_handlers_map = {
        'add': add_contact_handler,
        'add-email': add_email_handler,
        'change-email': change_email_handler,
        'set-address': set_address,
        'change': change_phone_handler,
        'phone': get_contact_handler,
        'all': get_all_contacts_handler,
        'add-birthday': add_birthday_handler,
        'show-birthday': get_birthday_handler,
        'birthdays': get_all_birthdays_handler,
        'upcoming-birthdays': get_upcoming_birthdays,
        'search-contacts': search_contacts,
        'delete-contact': delete_contact,
    }

    notes_handlers_map = {
        'notes': get_all_notes_handler,
        'add-note': add_note_handler,
        'edit-note': edit_note_handler,
        'delete-note': delete_note_handler,
        'search-note': search_note_handler,
        'add-note-tag': add_tag_to_note_handler,
        'notes-by-tags': sort_notes_by_tags,
    }

    if command in no_args_handlers_map:
        return no_args_handlers_map[command]()

    if command in contact_handlers_map:
        return contact_handlers_map[command](assistant.address_book, *arguments)

    if command in notes_handlers_map:
        return notes_handlers_map[command](assistant.notebook, *arguments)

    raise KeyError(f'Command {command} was not found')


def hello_handler():
    return 'Hello, how can I help you?'


def help_handler():
    return f'''Possible commands:
{Fore.LIGHTWHITE_EX}{Back.BLUE}help{Style.RESET_ALL} - prints list of available commands
{Fore.LIGHTWHITE_EX}{Back.BLUE}hello{Style.RESET_ALL} - prints a greeting 
{Fore.LIGHTWHITE_EX}{Back.BLUE}add [name] [phone number]{Style.RESET_ALL} - create a contact with a phone number
{Fore.LIGHTWHITE_EX}{Back.BLUE}add-email [name] [email] [is_primary]{Style.RESET_ALL} - Add or change email in contact, or create new contact with email if not exists
{Fore.LIGHTWHITE_EX}{Back.BLUE}change-email [name] [old_email_value] [new_email_value] [is_primary]{Style.RESET_ALL} - Add or change email in contact, or create new contact with email if not exists
{Fore.LIGHTWHITE_EX}{Back.BLUE}set-address [name] [address]{Style.RESET_ALL} - set address into contact
{Fore.LIGHTWHITE_EX}{Back.BLUE}change [name] [phone number]{Style.RESET_ALL} - changes a contact phone number 
{Fore.LIGHTWHITE_EX}{Back.BLUE}delete-contact [name] {Style.RESET_ALL} - delete contact
{Fore.LIGHTWHITE_EX}{Back.BLUE}phone [name]{Style.RESET_ALL} - prints contacts phone number
{Fore.LIGHTWHITE_EX}{Back.BLUE}all{Style.RESET_ALL} - prints all contacts
{Fore.LIGHTWHITE_EX}{Back.BLUE}search-contacts [search_term]{Style.RESET_ALL} - search contacts by names or phones
{Fore.LIGHTWHITE_EX}{Back.BLUE}add-birthday [name] [birthday]{Style.RESET_ALL} - adds birthday to a contact
{Fore.LIGHTWHITE_EX}{Back.BLUE}show-birthday [name]{Style.RESET_ALL} - prints contact's birthday
{Fore.LIGHTWHITE_EX}{Back.BLUE}birthdays{Style.RESET_ALL} - prints all birthdays
{Fore.LIGHTWHITE_EX}{Back.BLUE}upcoming-birthdays [number of days from today]{Style.RESET_ALL} - search for contacts on their birthday within a specified number of days from today
{Fore.LIGHTWHITE_EX}{Back.BLUE}notes{Style.RESET_ALL} - prints all notes
{Fore.LIGHTWHITE_EX}{Back.BLUE}notes-by-tags{Style.RESET_ALL} - prints all notes sorted by tags
{Fore.LIGHTWHITE_EX}{Back.BLUE}add-note [title] [text note]{Style.RESET_ALL} - adds a new note and prints it's ID
{Fore.LIGHTWHITE_EX}{Back.BLUE}edit-note [ID] [text note]{Style.RESET_ALL} - updates note with given ID
{Fore.LIGHTWHITE_EX}{Back.BLUE}delete-note [ID]{Style.RESET_ALL} - removes note with given ID
{Fore.LIGHTWHITE_EX}{Back.BLUE}search-note [text]{Style.RESET_ALL} - searches for notes by search phrase
{Fore.LIGHTWHITE_EX}{Back.BLUE}add-note-tag [ID] [tag]{Style.RESET_ALL} - adds a tag to a note
{Fore.LIGHTWHITE_EX}{Back.BLUE}close{Style.RESET_ALL} або {Fore.YELLOW}{Back.BLUE}exit{Style.RESET_ALL} - terminates a program
    '''


@input_error
def add_contact_handler(contacts: AddressBook, name: str, phone: str):
    name = name.lower().capitalize()

    try:
        record = contacts.find_by_name(name)
    except ValueError:
        record = Record(name)

    record.add_phone(phone)
    contacts[name] = record

    return format_success('Contact added')


@input_error
def add_email_handler(contacts: AddressBook, name: str, email: str, is_primary: str = "False"):
    name = name.lower().capitalize()
    primary = parser_bool_from_str(is_primary)

    record = contacts.find_by_name(name)
    if record is None:
        record = Record(name)

    record.add_email(email, primary)
    contacts[name] = record

    return format_success('Email added')


@input_error
def change_phone_handler(contacts: AddressBook, name: str, phone: str):
    name = name.lower().capitalize()
    try:
        record = contacts.find_by_name(name)
        record.edit_phone(record.phones[0], phone)
        contacts[name] = record

        return format_success('Contact updated.')
    except ValueError:
        return format_error('Contact not found.')


@input_error
def change_email_handler(contacts: AddressBook, name: str, old_email: str, new_email: str, is_primary: str):
    name = name.lower().capitalize()
    primary = parser_bool_from_str(is_primary)

    try:
        record = contacts.find_by_name(name)
        record.change_email(old_email, new_email, primary)
        contacts[name] = record

        return format_success('Email updated.')
    except ValueError as error:
        return format_error(error)


@input_error
def set_address(contacts: AddressBook, name: str, *args):
    try:
        record = contacts.find_by_name(name)
        record.set_address(' '.join(args))

        return format_success('The address is set.')
    except ValueError as error:
        return format_error(error)


@input_error
def get_contact_handler(contacts: AddressBook, name: str, *args):
    try:
        return contacts.find_by_name(name.lower().capitalize())
    except ValueError:
        return format_error('Contact not found.')


def get_all_contacts_handler(contacts: AddressBook, *args):
    if len(contacts) == 0:
        return format_error('AddressBook is empty.')
    return __records_to_str(contacts.values())


def add_birthday_handler(contacts: AddressBook, name: str, birthday: str):
    try:
        contacts.find_by_name(name.lower().capitalize()).add_birthday(birthday)
        return format_success('Contact updated.')
    except ValueError as error:
        return format_error(error)


def search_contacts(contacts: AddressBook, search_term: str) -> str:
    records = contacts.search(search_term)
    if len(records) == 0:
        return format_error('Contacts not found.')
    return __records_to_str(records)


def delete_contact(contacts: AddressBook, name: str) -> str:
    try:
        contacts.delete(name)
        return format_success('Contact deleted.')
    except ValueError as error:
        return format_error(error)


def get_birthday_handler(contacts: AddressBook, name: str, *args):
    try:
        return contacts.find_by_name(name.lower().capitalize()).birthday
    except ValueError:
        return format_error('Contact not found.')


def get_upcoming_birthdays(contacts: AddressBook, coming_days: str):
    records = contacts.get_upcoming_birthdays(int(coming_days))
    if len(records) == 0:
        return format_error('Contacts not found.')
    return __records_to_str(records)


def get_all_birthdays_handler(contacts: AddressBook, *args):
    return format_success('\n'.join(map(lambda name: f'{name}: {contacts[name].birthday}', contacts.keys())))


def get_all_notes_handler(notes: Notebook, *args):
    return format_success('\n\n\n'.join(map(lambda note_id: f'Note[{note_id}]\n{notes[note_id]}', notes.keys())))


def add_note_handler(notes: Notebook, title: str, text: str, *args):
    return format_success(f'Note added with ID: {notes.add_note(title, text)}')


def edit_note_handler(notes: Notebook, note_id: str, text: str, *args):
    try:
        notes.edit_note(note_id, text)
    except IndexError:
        return format_error('Note not found.')

    return format_success(f'Note #{note_id} updated')


def delete_note_handler(notes: Notebook, note_id: str, *args):
    try:
        notes.delete_note(note_id)
        return format_success(f'Note #{note_id} deleted')
    except IndexError:
        return format_error('Note not found.')


def search_note_handler(notes: Notebook, phrase: str, *args):
    notes = notes.find_notes(phrase)
    return format_success('\n'.join(map(lambda note: f'Note: "{note}"', notes)))


def add_tag_to_note_handler(notes: Notebook, note_id: str, tag: str, *args):
    try:
        notes.add_tag(note_id, tag)
        return format_success(f'Note #{note_id} updated')
    except IndexError:
        return format_error('Note not found.')


def sort_notes_by_tags(notes: Notebook, *args):
    return_text = ''
    for tag, notes in notes.sort_notes_by_tags().items():
        return_text += format_success(f'Tag: "{tag}", Notes:\n')
        for note_id, note in notes.items():
            return_text += format_success(f'Note[{note_id}]\n{note}')
        return_text += format_success(f'\n{'='*10}\n')

    return return_text


def format_notes_dict(notes):
    return format_success('\n'.join(map(lambda note_id: f'ID[{note_id}] Note: "{notes[note_id]}"', notes.keys())))


def parser_bool_from_str(val: str) -> bool:
    return val.lower() == 'true'


def __records_to_str(records: iter) -> str:
    return '\n'.join(map(lambda record: f'{record}', records))



from collections import UserDict

from .command import Command


class CommandRegistry(UserDict):
    def __init__(self):
        super().__init__()

    def register(self, command: Command):
        self.data[command.name] = command

    def list(self) -> list[str]:
        return list(self.data.keys())




from re import match
from colorama import Fore, Back, Style

from taipan_assistant.contacts.record.field import Field


class Email(Field):

    def __init__(self, value: str, primary: bool):
        if not match(r'^.+@.+\..+$', value):
            raise ValueError(
                'Email should be in format {one or more any char}@{one or more any char}.{one or more any char}'
            )

        self.primary = primary
        super().__init__(value=value)

    def __str__(self):
        if self.primary:
            return f'{Fore.LIGHTWHITE_EX}{Back.GREEN}{super().__str__()}{Style.RESET_ALL}'

        return f'{super().__str__()}'



from taipan_assistant.contacts.record.address import Address
from taipan_assistant.contacts.record.birthday import Birthday
from taipan_assistant.contacts.record.email import Email
from taipan_assistant.contacts.record.name import Name
from taipan_assistant.contacts.record.phone import Phone, PhoneCollection


class Record:
    def __init__(self, name):
        self.name = Name(name)
        self.phones = PhoneCollection()
        self.birthday = None
        self.emails = []
        self.address = None

    def add_phone(self, phone: str):
        for p in self.phones:
            if p.value == phone:
                return

        self.phones.append(Phone(phone))

    def set_address(self, address: str):
        self.address = Address(address)

    def add_email(self, email: str, is_primary: bool):
        if not self.__can_update_primary_email(email, is_primary):
            raise ValueError("If a contact has an email address, there must be one primary email address")

        # if new is primary set all existed emails as secondary
        if is_primary:
            self.__set_all_emails_as_secondary()

        # if email already there is - update primary status
        for e in self.emails:
            if e.value == email:
                e.primary = is_primary

                return

        if len(self.emails) == 0:
            is_primary = True

        self.emails.append(Email(email, is_primary))

    def change_email(self, old_email: str, new_email: str, is_primary: bool):
        if not self.__can_update_primary_email(old_email, is_primary):
            raise ValueError("If a contact has an email address, there must be one primary email address")

        email = self.find_email(old_email)

        # if new is primary set all existed emails as secondary
        if is_primary:
            self.__set_all_emails_as_secondary()

        email.value = new_email
        email.primary = is_primary

    def __can_update_primary_email(self, email_for_update: str, new_primary_val: bool) -> bool:
        if new_primary_val:
            return True

        # We can`t change primary to not primary, because primary mail is required
        for e in self.emails:
            if e.value == email_for_update and e.primary:
                return False

        return True

    def __set_all_emails_as_secondary(self):
        for e in self.emails:
            e.primary = False

    def find_email(self, email: str) -> Email:
        for e in self.emails:
            if e.value == email:
                return e

        raise ValueError(f'Email {email} does not belong record {self.name}')

    def add_birthday(self, birthday: str):
        self.birthday = Birthday(birthday)

    def remove_phone(self, phone: str):
        phone_field = Phone(phone)
        if phone_field not in self.phones:
            raise ValueError(f'Phone number {phone} does not belong record {self.name}')

        self.phones.remove(phone_field)

    def edit_phone(self, old_phone: str, new_phone: str):
        for phone in self.phones:
            if phone.value == old_phone:
                phone.value = new_phone
                break

    def find_phone(self, phone: str) -> Phone:
        for phone_field in self.phones:
            if phone_field.value == phone:
                return phone_field

        raise ValueError(f'Phone number {phone} does not belong record {self.name}')

    def __str__(self) -> str:
        result_str = (
            f"Contact name: {self.name.value}, "
            f"phones: {'; '.join(p.value for p in self.phones)}"
        )

        if len(self.emails) > 0:
            result_str = result_str + f", emails: {'; '.join(str(e) for e in self.emails)}"

        if self.address is not None:
            result_str = result_str + f", address: {self.address.value}"

        if self.birthday is not None:
            result_str = result_str + f", birthday: {self.birthday}"

        return result_str



from collections import UserList
from re import match
from taipan_assistant.contacts.record.field import Field


class Phone(Field):
    # реалізація класу
    def __init__(self, value):
        if not match(r'^\d{10}$', value):
            raise ValueError('Phone number must be 10 digits')

        super().__init__(value=value)


class PhoneCollection(UserList):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return ' '.join(list(map(str, self.data)))



from typing import Callable
from functools import wraps

from .output_formatter import format_error


def input_error(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyError as e:
            return format_error(f'Invalid command. [KeyError {e}]')
        except ValueError as e:
            return format_error(f'Enter the argument for the command [ValueError {e}]')
        except IndexError as e:
            return format_error(f'Enter the argument for the command [IndexError {e}]')
        except TypeError as e:
            return format_error(f'Enter the argument for the command [TypeError {e}]')

    return wrapper



from collections import UserDict, defaultdict

import ngram


class Engine(UserDict):
    """
    This class allows to index and search documents. Result of the search is sorted by the best match.

    Example:
        class Person:
            def __init__(self, name, email):
                self.name = name
                self.email = email

        engine = Engine(["name", "email"])
        engine.add_to_index(Person("John Doe", "findmeinthemorgue@gmail.com"))
        engine.add_to_index(Person("Jane Doe", "findmeinnearjohn@gmail.com"))

        result = engine.search("one")

    :param fields: list of fields to index and search by. Those fields MUST support casting to string
    :param search_threshold: allows to make search more or less strict
    """
    def __init__(self, fields: list, search_threshold: float = 0.2):
        super().__init__()

        self.__fields = fields
        self.__search_threshold = search_threshold

        self.__rebuild_index()

    def add_to_index(self, entity_id, entity):
        """
        :param entity_id: id of the entity
        :param entity: object to be indexed
        :return:
        """
        self.data[entity_id] = entity

        self.__rebuild_index()

    def remove_from_index(self, entity_id):
        del self.data[entity_id]

        self.__rebuild_index()

    def find(self, search_string: str) -> list:
        result = []
        found_items = {}

        keys = []
        for key in self.__index.keys():
            keys.append(str(key))

        ng = ngram.NGram(keys, N=2)
        for value, weight in ng.search(search_string):
            if weight < self.__search_threshold:
                continue

            for entity_id in self.__index[value]:
                if entity_id in found_items:
                    continue

                result.append(self.data[entity_id])
                found_items[entity_id] = True

        return result

    def __rebuild_index(self):
        self.__index = defaultdict(set)

        for entity_id, entity in self.data.items():
            for field in self.__fields:
                field_value = getattr(entity, field)

                self.__index[str(field_value)].add(entity_id)


from ..contacts.address_book import AddressBook
from ..notebook.notebook import Notebook


class Assistant:
    def __init__(self):
        self.address_book = AddressBook()
        self.notebook = Notebook()



from datetime import datetime
from taipan_assistant.contacts.record.field import Field


class Birthday(Field):
    def __init__(self, value):
        try:
            super().__init__(value=datetime.strptime(value, '%d.%m.%Y'))
        except ValueError:
            raise ValueError("Invalid date format. Use DD.MM.YYYY")

    def __str__(self):
        return self.value.strftime("%d.%m.%Y")



from collections import UserDict
from datetime import datetime

from .record.birthday import Birthday
from .record.record import Record
from ..search.engine import Engine


class AddressBook(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_engine = Engine(['name', 'phones'])

    def __setitem__(self, key, item):
        super().__setitem__(key, item)
        self.__add_record_to_index(item)

    def __delitem__(self, key):
        super().__delitem__(key)
        self.search_engine.remove_from_index(key)

    def add_record(self, record: Record):
        self.data[str(record.name)] = record
        self.__add_record_to_index(record)

    def find_by_name(self, name: str) -> Record:
        if name not in self.data.keys():
            raise ValueError('Record not found')

        return self.data[name]

    def search(self, search_term: str) -> list:
        return self.search_engine.find(search_term)

    def delete(self, name: str):
        if name not in self.data.keys():
            raise ValueError('Record not found')

        record = self.data.pop(name)
        self.search_engine.remove_from_index(record.name)


    def get_upcoming_birthdays(self, coming_days: int) -> list:
        """Find users who have a birthday in the next coming_days days"""

        result = []

        for record in self.data.values():
            if not type(record.birthday) == Birthday:
                continue

            today = datetime.today()

            current_year = today.year
            year_diff = current_year - record.birthday.value.year
            # find the nearest next birthday
            next_birthday = record.birthday.value.replace(year=record.birthday.value.year + year_diff).date()
            if next_birthday < today.date():
                next_birthday = next_birthday.replace(year=next_birthday.year + 1)

            birthday_diff = next_birthday - today.date()
            # skip user if the birthday is more than coming_days days away
            if birthday_diff.days > coming_days:
                continue
            record.congratulation_date = next_birthday.strftime("%Y.%m.%d")
            result.append(record)

        return result

    def __add_record_to_index(self, record: Record):
        self.search_engine.add_to_index(record.name, record)



from typing import Callable


class Command:
    def __init__(self, name: str, handler: Callable, description: str = None):
        self.name = name
        self.__handler = handler
        self.description = description

    def __call__(self, *args, **kwargs):
        return self.__handler(*args, **kwargs)



from colorama import Fore, Back, Style


def format_error(error: str):
    return f'{Fore.LIGHTWHITE_EX}{Back.RED}{error}{Style.RESET_ALL}'


def format_success(message: str):
    return f'{Fore.GREEN}{message}{Style.RESET_ALL}'
